"""
Logging utilities, console handlers, and rotational file handlers.
"""

import logging
from logging.handlers import TimedRotatingFileHandler
import sys

from websnap.validators import LogConfigModel
from websnap.constants import LogFormatter, LogLevel


def get_log_level(log_level: str = "INFO") -> int:
    """
    Return logging level for logger. Default log level: logging.INFO

    Args:
        log_level: logging level represented as upper case string

    Returns:
        Integer that corresponds to logging level.
    """
    match log_level:
        case LogLevel.DEBUG.value:
            level = logging.DEBUG
        case LogLevel.WARNING.value:
            level = logging.WARNING
        case LogLevel.ERROR.value:
            level = logging.ERROR
        case LogLevel.CRITICAL.value:
            level = logging.CRITICAL
        case LogLevel.INFO.value | _:
            level = logging.INFO

    return level


def get_console_handler() -> logging.StreamHandler:
    """
    Return formatted console handler.
    """
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(LogFormatter.formatter())
    return console_handler


def get_file_handler(
    filename: str, when: str, interval: int, backup_count: int
) -> logging.handlers.TimedRotatingFileHandler:
    """
    Return formatted rotational file handler for logs.
    To learn more about logging TimedRotating FileHandler and the values acceptable for
     'when' and how to use interval and backupCount see:
    https://docs.python.org/3/library/logging.handlers.html#timedrotatingfilehandler

    Args:
        filename: Name of log file.
        when: Value used to assign when logs rotate.
        interval: How frequently to rotate logs.
        backup_count: If more than 0 then at most <backup_count> files will be kept.
    """
    file_handler = TimedRotatingFileHandler(
        filename=filename, when=when, interval=interval, backupCount=backup_count
    )
    file_handler.setFormatter(LogFormatter.formatter())
    return file_handler


def get_custom_logger(
    name: str,
    level: str = "INFO",
    file_logs: bool = False,
    config: LogConfigModel | None = None,
) -> logging.getLogger:
    """
    Return logger with console handler and optional file handler.
    Default logging level is 'INFO'.

    Args:
        name: Name of logger.
        level: Logging level represented as string.
        file_logs: If True then implements rotating file logs.
        config: Validated log config_templates.
    """
    try:
        _loglevel = level.upper()
    except AttributeError:
        raise Exception("Argument loglevel must be a string")

    logger = logging.getLogger(name)
    logger.setLevel(get_log_level(_loglevel))
    logger.addHandler(get_console_handler())

    if file_logs:
        logger.addHandler(
            get_file_handler(
                filename=f"{name}.log",
                when=config.log_when,
                interval=config.log_interval,
                backup_count=config.log_backup_count,
            )
        )

    logger.propagate = False

    return logger



"""
Function websnap() downloads files from URLs and uploads them to S3 bucket.
Also supports writing downloaded files to local machine.
"""

import configparser

from math import ceil
import time

from websnap.validators import (
    get_config_parser,
    validate_log_config,
    validate_s3_config,
    validate_min_size_kb,
    S3ConfigModel,
    LogConfigModel,
)
from websnap.logger import get_custom_logger
from websnap.logic import write_urls_locally, write_urls_to_s3

__all__ = ["websnap"]

LOGGER_NAME = "websnap"


def websnap(
    config: str = "config.ini",
    log_level: str = "INFO",
    file_logs: bool = False,
    s3_uploader: bool = False,
    backup_s3_count: int | None = None,
    early_exit: bool = False,
    repeat_minutes: int | None = None,
) -> None | Exception:
    """
    Copies files hosted at URLs in config_templates and then uploads them
    to S3 bucket or local machine.
    Optionally customize rotating logs.
    Optionally repeat websnap file processing iteration.

    Args:
        config: Path to ini config file.
        log_level: Level to use for logging.
        file_logs: If True then implements rotating file logs.
        s3_uploader: If True then uploads files to S3 bucket.
        backup_s3_count: Copy and backup S3 objects in each config_templates section
            <backup_s3_count> times,
            remove object with the oldest last modified timestamp.
            If omitted then default value is None and objects are not copied.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
        repeat_minutes: Run websnap continuously every <repeat> minutes, if omitted
            then default value is None and websnap will not repeat.
    """
    # Validate log settings in config_templates and setup log
    try:
        conf_parser = get_config_parser(config)
        if not isinstance(conf_parser, configparser.ConfigParser):
            raise Exception(conf_parser)

        conf_log = validate_log_config(conf_parser)
        if not isinstance(conf_log, LogConfigModel):
            raise Exception(conf_log)

        log = get_custom_logger(
            name=LOGGER_NAME,
            level=log_level,
            file_logs=file_logs,
            config=conf_log,
        )
    except Exception as e:
        raise Exception(e)

    # Validate min_size_kb in config_templates
    min_size_kb = validate_min_size_kb(conf_parser)
    if not isinstance(min_size_kb, int):
        raise Exception(min_size_kb)

    # Download and write URL files
    is_repeat = True
    while is_repeat:

        # Do not repeat iteration if repeat_minutes is None
        is_repeat = repeat_minutes is not None

        start_time = time.time()

        log.info("******* STARTED WEBSNAP ITERATION *******")
        log.info(
            f"Read config_templates file: '{config}', it has sections: "
            f"{conf_parser.sections()}"
        )

        if s3_uploader:
            conf_s3 = validate_s3_config(conf_parser)
            if not isinstance(conf_s3, S3ConfigModel):
                raise Exception(conf_s3)
            write_urls_to_s3(
                conf_parser, conf_s3, log, min_size_kb, backup_s3_count, early_exit
            )
        else:
            write_urls_locally(conf_parser, log, min_size_kb, early_exit)

        log.info("Finished websnap iteration")
        exec_time = int(time.time() - start_time)

        if is_repeat:
            interval_seconds = int(repeat_minutes) * 60
            if interval_seconds > exec_time:
                wait_seconds = interval_seconds - exec_time
                wait_minutes = wait_seconds / 60
                log.info(
                    f"Sleeping {wait_seconds} seconds (about {ceil(wait_minutes)} "
                    f"minutes) before next iteration..."
                )
                time.sleep(wait_seconds)

    return



"""Config utilities, parses and validates config_templates .ini files."""

import configparser
from pathlib import Path
from pydantic import (
    BaseModel,
    ValidationError,
    PositiveInt,
    AnyHttpUrl,
    AnyUrl,
    field_validator,
    NonNegativeInt,
)
from typing import Optional

from websnap.constants import LogRotation, MIN_SIZE_KB


def get_config_parser(config_path: str) -> configparser.ConfigParser | Exception:
    """
    Return ConfigParser object.
    Returns Exception if fails.

    Args:
        config_path (str): Path to config_templates.ini file.
    """
    try:
        config_file = Path(config_path)
        config = configparser.ConfigParser()
        conf = config.read(config_file)

        if not conf:
            return Exception(f"File {config_path} not found")

        if len(config.sections()) < 1:
            return Exception(
                f"File '{config_path}' not valid config_templates, "
                f"does not have any sections"
            )

        return config

    except Exception as e:
        return Exception(f"{e}")


class LogConfigModel(BaseModel):
    """
    Class with required log config_templates values and their types.
    """

    log_when: str
    log_interval: PositiveInt
    log_backup_count: NonNegativeInt


def validate_log_config(
    config_parser: configparser.ConfigParser,
) -> LogConfigModel | Exception:
    """
    Return LogConfigModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser (configparser.ConfigParser): ConfigParser object
    """
    try:
        log = {
            "log_when": config_parser.get(
                "DEFAULT", "log_when", fallback=LogRotation.WHEN.value
            ),
            "log_interval": config_parser.get(
                "DEFAULT", "log_interval", fallback=LogRotation.INTERVAL.value
            ),
            "log_backup_count": config_parser.getint(
                "DEFAULT", "log_backup_count", fallback=LogRotation.BACKUP_COUNT.value
            ),
        }
        return LogConfigModel(**log)
    except ValidationError as e:
        return Exception(f"Failed to validate config_templates, error(s): {e}")
    except ValueError as e:
        return Exception(
            f"Incorrect log related value in config_templates, error(s): {e}"
        )
    except Exception as e:
        return Exception(f"{e}")


def validate_min_size_kb(config_parser: configparser.ConfigParser) -> int | Exception:
    """
    Return min_size_kb from config_templates as integer.
    Returns Exception if validation fails.

    Args:
        config_parser: ConfigParser object
    """
    try:
        min_size_kb = config_parser.getint(
            "DEFAULT", "min_size_kb", fallback=MIN_SIZE_KB
        )
        if min_size_kb >= 0:
            return min_size_kb
        else:
            raise ValueError(
                "Value for config_templates value 'min_size_kb' must be greater than 0"
            )
    except ValidationError as e:
        return Exception(
            f"Failed to validate config_templates value 'min_size_kb, error(s): {e}"
        )
    except ValueError as e:
        return Exception(
            f"Incorrect value for config_templates value 'min_size_kb', error(s): {e}"
        )
    except Exception as e:
        return Exception(f"{e}")


class ConfigSectionModel(BaseModel):
    """
    Class with required config_templates section values (for writing to local machine).
    """

    url: AnyHttpUrl
    file_name: str
    directory: Optional[str] = None


def validate_config_section(
    config_parser: configparser.ConfigParser, section: str
) -> ConfigSectionModel | Exception:
    """
    Return ConfigSectionModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser: ConfigParser object
        section: Name of section being validated
    """
    try:
        conf_section = {
            "url": config_parser.get(section, "url"),
            "file_name": config_parser.get(section, "file_name"),
        }
        if directory := config_parser.get(section, "directory", fallback=None):
            conf_section["directory"] = directory
        return ConfigSectionModel(**conf_section)
    except ValidationError as e:
        return Exception(
            f"Failed to validate config_templates section '{section}', error(s): {e}"
        )
    except ValueError as e:
        return Exception(
            f"Incorrect value in config_templates section '{section}', error(s): {e}"
        )
    except Exception as e:
        return Exception(f"{e}")


class S3ConfigModel(BaseModel):
    """
    Class with required S3 config_templates values and their types.
    """

    endpoint_url: AnyUrl
    aws_access_key_id: str
    aws_secret_access_key: str


def validate_s3_config(
    config_parser: configparser.ConfigParser,
) -> S3ConfigModel | Exception:
    """
    Return S3ConfigModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser (configparser.ConfigParser): ConfigParser object
    """
    try:
        s3_conf = {
            "endpoint_url": config_parser.get("DEFAULT", "endpoint_url"),
            "aws_access_key_id": config_parser.get("DEFAULT", "aws_access_key_id"),
            "aws_secret_access_key": config_parser.get(
                "DEFAULT", "aws_secret_access_key"
            ),
        }
        return S3ConfigModel(**s3_conf)
    except ValidationError as e:
        return Exception(f"Failed to validate S3 config_templates, error(s): {e}")
    except ValueError as e:
        return Exception(f"Incorrect value in S3 config_templates, error(s): {e}")
    except Exception as e:
        return Exception(e)


class S3ConfigSectionModel(BaseModel):
    """
    Class with required config_templates section values (for writing to S3 bucket).
    """

    url: AnyUrl
    bucket: str
    key: str

    @field_validator("key")
    @classmethod
    def key_must_contain_period(cls, v: str) -> str:
        key_split = v.rpartition(".")
        if not key_split[1]:
            raise ValueError("Config section key requires a file extension")
        return v

    @field_validator("key")
    @classmethod
    def key_must_not_start_with_slash(cls, v: str) -> str:
        if v.startswith("/"):
            raise ValueError("Config section key cannot start with a '/'")
        return v


def validate_s3_config_section(
    config_parser: configparser.ConfigParser, section: str
) -> S3ConfigSectionModel | Exception:
    """
    Return S3ConfigSectionModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser: ConfigParser object
        section: Name of section being validated
    """
    try:
        conf_section = {
            "url": config_parser.get(section, "url"),
            "bucket": config_parser.get(section, "bucket"),
            "key": config_parser.get(section, "key"),
        }
        return S3ConfigSectionModel(**conf_section)
    except ValidationError as e:
        return Exception(
            f"Failed to validate config_templates section '{section}', error(s): {e}"
        )
    except ValueError as e:
        return Exception(
            f"Incorrect value in config_templates section '{section}', error(s): {e}"
        )
    except Exception as e:
        return Exception(f"{e}")



"""Enums and constants."""

import logging
from enum import Enum

MIN_SIZE_KB: int = 0


class LogFormatter(Enum):
    """Class with values used to format logs."""

    FORMAT = (
        "%(asctime)s | %(levelname)s | %(module)s.%(funcName)s:%(lineno)d | %(message)s"
    )
    DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    @classmethod
    def formatter(cls):
        return logging.Formatter(fmt=cls.FORMAT.value, datefmt=cls.DATE_FORMAT.value)


class LogLevel(Enum):
    """Class with supported log levels."""

    DEBUG = "DEBUG"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"
    INFO = "INFO"


class LogRotation(Enum):
    """Class with default values used by rotating file logs."""

    WHEN = "D"
    INTERVAL = 1
    BACKUP_COUNT = 0



"""
Supporting functions used to download and write files hosted
at URLs to S3 bucket or local machine.
"""

import configparser
import logging
import os

import requests
import boto3
from botocore.exceptions import ClientError
import sys

from websnap.validators import (
    validate_config_section,
    S3ConfigModel,
    validate_s3_config_section,
    ConfigSectionModel,
    S3ConfigSectionModel,
)


def terminate_program(early_exit: bool):
    """Terminates program execution if argument early_exit is True."""
    if early_exit:
        sys.exit("Error occurred: check logs for details")
    return


def get_url_content(
    url: str, section: str, log: logging.getLogger, early_exit: bool = False
) -> bytes | None:
    """
    Return content of response from HTTP GET request.
    If response status code is >= 400 then terminate program if argument early_exit
    is True, else return None.

    Args:
        url: URL to download.
        section: Name of config_templates section being processed.
        log: Logger object created with customized configuration file.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    response = requests.get(url)

    if not response.ok:
        log.error(
            f"Config section '{section}': "
            f"URL returned unsuccessful HTTP response "
            f"status code {response.status_code}"
        )
        terminate_program(early_exit)
        return None

    return response.content


def is_min_size_kb(
    url_content: bytes,
    min_size_kb: int,
    section: str,
    log: logging.getLogger,
    early_exit: bool = False,
) -> bool | None:
    """
    Return True if url_content is greater than min_size_kb.
    Else return False or terminate program (if argument early_exit is True).

    Args:
        url_content: Content of response from HTTP request.
        min_size_kb: Minimum threshold in kilobytes that URL response content must be to
            write or upload file.
        section: Name of config_templates section being processed.
        log: Logger object created with customized configuration file.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    data_kb = url_content.__sizeof__() / 1024

    if data_kb < min_size_kb:
        log.error(
            f"Config section '{section}': "
            f"URL response content in config_templates section {section} is less than "
            f"config_templates value 'min_size_kb' {min_size_kb}"
        )
        terminate_program(early_exit)
        return False

    return True


def write_urls_locally(
    conf_parser: configparser.ConfigParser,
    log: logging.getLogger,
    min_size_kb: int,
    early_exit: bool = False,
):
    """
    Download files hosted at URLS in config_templates and then write them to local
    machine.

    Args:
        conf_parser: ConfigParser object created from parsing configuration file.
        log: Logger object created with customized configuration file.
        min_size_kb: Minimum threshold in kilobytes that URL response content must be to
            write file.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    for section in conf_parser.sections():

        try:
            conf = validate_config_section(conf_parser, section)
            if not isinstance(conf, ConfigSectionModel):
                log.error(f"Config section '{section}': {conf}")
                terminate_program(early_exit)
                continue

            if conf.directory and not os.path.isdir(conf.directory):
                log.error(
                    f"Config section '{section}': directory '{conf.directory}' "
                    f"does not exist"
                )
                terminate_program(early_exit)
                continue

            url_content = get_url_content(str(conf.url), section, log, early_exit)
            if not url_content:
                continue

            is_min_size = is_min_size_kb(
                url_content, min_size_kb, section, log, early_exit
            )
            if not is_min_size:
                continue

            if conf.directory:
                file_path = f"{conf.directory}/{conf.file_name}"
            else:
                file_path = f"{conf.file_name}"

            with open(file_path, "wb") as f:
                f.write(url_content)
                log.info(
                    f"Successfully downloaded URL content and wrote file locally in "
                    f"config_templates section: {section}"
                )

        except Exception as e:
            log.error(f"Config section '{section}', error(s): {e}")
            terminate_program(early_exit)

    return


def copy_s3_object(
    client: boto3.Session.client,
    conf: S3ConfigSectionModel,
    log: logging.getLogger,
    section: str,
    early_exit: bool = False,
):
    """
    Copy an object using S3 object config_templates.

    New object's name is constructed using the 'LastModified' timestamp of original
    object.

    Args:
        client : boto3.Session.client object created using configuration file values.
        conf: S3ConfigSectionModel object created from validated
            section of configuration file.
        log: Logger object created with customized configuration file.
        section: Name of config_templates section being processed.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """

    try:
        obj = client.head_object(Bucket=conf.bucket, Key=conf.key)

        last_modified = obj.get("LastModified")
        format_date = "%Y-%m-%d_%H-%M-%S"
        datetime_str = last_modified.strftime(format_date)
        key_split = conf.key.rpartition(".")
        key_copy = f"{key_split[0]}_{datetime_str}{key_split[1]}{key_split[2]}"

        response = client.copy_object(
            CopySource={"Bucket": conf.bucket, "Key": conf.key},
            Bucket=conf.bucket,
            Key=key_copy,
        )

        status_code = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

        if status_code == 200:
            log.info(
                f"S3 config_templates section '{section}': "
                f"Created new backup file '{key_copy}'"
            )
        else:
            log.error(
                f"S3 config_templates section '{section}': "
                f"Object backup attempt returned "
                f"unexpected HTTP response {status_code}"
            )
            terminate_program(early_exit)

    except ClientError as e:
        log.error(e)
        terminate_program(early_exit)

    return


def delete_s3_backup_object(
    client: boto3.Session.client,
    conf: S3ConfigSectionModel,
    log: logging.getLogger,
    section: str,
    backup_s3_count: int,
    early_exit: bool = False,
):
    """
    Delete a S3 backup object using S3 object config_templates.
    Only deletes object if backup objects exceed backup_s3_count.

    Only deletes object that corresponds to the file name in the configured key,
    allows for a timestamp in key created using copy_s3_object().

    Args:
        client : boto3.Session.client object created using configuration file values.
        conf: S3ConfigSectionModel object created from validated
            section of configuration file.
        log: Logger object created with customized configuration file.
        section: Name of config_templates section being processed.
        backup_s3_count: Copy and backup S3 objects in config_templates
            <backup_s3_count> times, remove object with the oldest last modified
            timestamp.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """

    try:
        key_split = conf.key.rpartition("/")

        if not key_split[0]:
            response = client.list_objects_v2(
                Bucket=conf.bucket,
            )
        else:
            response = client.list_objects_v2(
                Bucket=conf.bucket, Prefix=f"{key_split[0]}{key_split[1]}"
            )

        file_split = key_split[2].rpartition(".")
        file_start = f"{file_split[0]}_"
        file_end = f"{file_split[1]}{file_split[2]}"

        objs = [obj for obj in response.get("Contents")]
        match_objs = []

        for obj in objs:
            ky = obj.get("Key")
            ky_split = ky.rpartition("/")
            ky_file = ky_split[2]
            if ky_file.startswith(file_start) and ky_file.endswith(file_end):
                match_objs.append(obj)

        sorted_objs = sorted(match_objs, key=lambda x: x["LastModified"])

        if len(sorted_objs) > backup_s3_count:

            obj_oldest = sorted_objs[0]
            delete_key = obj_oldest.get("Key")

            resp = client.delete_object(Bucket=conf.bucket, Key=delete_key)

            status_code = resp.get("ResponseMetadata", {}).get("HTTPStatusCode")

            if status_code == 204:
                log.info(
                    f"S3 config_templates section '{section}': "
                    f"Deleted backup file '{delete_key}'"
                )
            else:
                log.error(
                    f"S3 config_templates section '{section}': Backup file delete "
                    f"attempt returned unexpected HTTP response {status_code}"
                )
                terminate_program(early_exit)

        else:
            log.info(
                f"S3 config_templates section '{section}': Current number of backup "
                f"files does not exceed backup S3 count {backup_s3_count}"
            )

    except ClientError as e:
        log.error(e)
        terminate_program(early_exit)

    return


def write_urls_to_s3(
    conf_parser: configparser.ConfigParser,
    conf_s3: S3ConfigModel,
    log: logging.getLogger,
    min_size_kb: int,
    backup_s3_count: int | None = None,
    early_exit: bool = False,
):
    """
    Download files hosted at URLS in config_templates and then upload them to S3 bucket.

    Args:
        conf_parser: ConfigParser object created from parsing configuration file.
        conf_s3: S3ConfigModel object created from validated configuration file.
        log: Logger object created with customized configuration file.
        min_size_kb: Minimum threshold in kilobytes that URL response content must be to
            upload file to S3 bucket.
        backup_s3_count: Copy and backup S3 objects in each config_templates section
            <backup_s3_count> times,
            remove object with the oldest last modified timestamp.
            If omitted then default value is None and objects are not copied or removed.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    session = boto3.Session(
        aws_access_key_id=conf_s3.aws_access_key_id,
        aws_secret_access_key=conf_s3.aws_secret_access_key,
    )

    client = session.client(service_name="s3", endpoint_url=str(conf_s3.endpoint_url))

    for section in conf_parser.sections():

        try:
            conf = validate_s3_config_section(conf_parser, section)
            if not isinstance(conf, S3ConfigSectionModel):
                log.error(f"Config section '{section}': {conf}")
                terminate_program(early_exit)
                continue

            url_content = get_url_content(str(conf.url), section, log, early_exit)
            if not url_content:
                continue

            is_min_size = is_min_size_kb(
                url_content, min_size_kb, section, log, early_exit
            )
            if not is_min_size:
                continue

            if backup_s3_count:
                copy_s3_object(client, conf, log, section, early_exit)
                delete_s3_backup_object(client, conf, log, section, backup_s3_count)

            response_s3 = client.put_object(
                Body=url_content, Bucket=conf.bucket, Key=conf.key
            )
            status_code = response_s3.get("ResponseMetadata", {}).get("HTTPStatusCode")

            if status_code == 200:
                log.info(
                    f"Config section '{section}': Successfully downloaded URL "
                    f"content and uploaded content to S3 object '{conf.key}'"
                )
            else:
                log.error(
                    f"Config section '{section}': S3 returned unexpected "
                    f"HTTP response {status_code}"
                )
                terminate_program(early_exit)

        except Exception as e:
            log.error(f"Config section '{section}', error(s): {e}")
            terminate_program(early_exit)

    return
